'use strict';

const str$1 = () => ('00000000000000000' + (Math.random() * 0xffffffffffffffff).toString(16)).slice(-16);
const uuid = () => {
    const a = str$1();
    const b = str$1();
    return a.slice(0, 8) + '-' + a.slice(8, 12) + '-4' + a.slice(13) + '-a' + b.slice(1, 4) + '-' + b.slice(4);
};
/**
 * Generates a random string using the UUID library.
 *
 * @returns A randomly generated string.
 */
const randomString = () => uuid();

/**
 * Compares the full text in a given data object with a search term.
 *
 * @template T - The generic type for the data object.
 * @param data - The data object to compare.
 * @param search - The search term to compare against the data object.
 * @param keys - The keys in the data object to compare against.
 * @returns - Returns true if the full text in the data object contains the search term, false otherwise.
 */
const compareFulltext = (data, search, ...keys) => {
    const target = String(search || "")
        .toLowerCase()
        .split(" ");
    const filterFn = (key) => {
        const source = String(data[key] || "")
            .toLowerCase()
            .split(" ");
        return target.every(function (term) {
            return source.some((word) => word.includes(term));
        });
    };
    let isOk = false;
    for (const key of keys) {
        isOk = isOk || filterFn(key);
    }
    return isOk;
};

/**
 * A function that compares two values and returns a numeric value based on their comparison.
 *
 * @param a - The first value to compare.
 * @param b - The second value to compare.
 * @returns - A numeric value indicating the result of the comparison.
 */
const compareFn = (a, b) => {
    if (typeof a === 'number' && typeof b === 'number') {
        return a - b;
    }
    else if (typeof a === 'boolean' && typeof b === 'boolean') {
        return (a ? 1 : 0) - (b ? 1 : 0);
    }
    else if (typeof a === 'string' && typeof b === 'string') {
        return a.localeCompare(b);
    }
    else {
        return 0;
    }
};
/**
 * Compares two arrays and determines if they are equal.
 *
 * @param a_arr - The first array to compare.
 * @param b_arr - The second array to compare.
 * @returns - Returns `true` if the arrays are equal, otherwise `false`.
 */
const compareArray = (a_arr, b_arr) => {
    if (Array.isArray(a_arr) && Array.isArray(b_arr)) {
        if (a_arr.length !== b_arr.length) {
            return false;
        }
        const a_sort = [...a_arr].sort(compareFn);
        const b_sort = [...b_arr].sort(compareFn);
        return a_sort.every((value, index) => value === b_sort[index]);
    }
    return false;
};

/**
 * Check if a given value is an object.
 *
 * @param obj - The value to check.
 * @returns - Returns `true` if the value is an object, else `false`.
 */
const isObject = (obj) => {
    if (typeof obj === 'object' && obj !== null) {
        return Object.getPrototypeOf(obj) === Object.prototype;
    }
    else {
        return false;
    }
};

/**
 * Formats a raw string using a template and optional parameters.
 *
 * @param raw - The raw string to be formatted.
 * @param template - The template string used for formatting.
 * @param [params] - Optional parameters for customization.
 * @param [params.symbol='0'] - The symbol used in the template to indicate characters to be replaced.
 * @param [params.allowed] - A function or regular expression used to filter characters in the raw string.
 * @param [params.replace] - A function used to replace characters in the raw string.
 * @returns The formatted string.
 */
const formatText = (raw, template, { symbol = '0', allowed, replace, } = {}) => {
    if (!template || !raw) {
        return raw;
    }
    if (replace) {
        const tempRaw = raw.split('');
        for (let i = 0; i !== raw.length; i++) {
            const pendingChar = replace(raw[i]) || null;
            if (pendingChar !== null) {
                tempRaw[i] = pendingChar;
            }
        }
        raw = tempRaw.join('');
    }
    if (allowed) {
        const pendingRemoveIdx = new Set();
        if (typeof allowed === 'function') {
            for (let i = 0; i !== raw.length; i++) {
                const lastChar = raw[i];
                if (!allowed(lastChar, i)) {
                    pendingRemoveIdx.add(i);
                }
            }
        }
        else {
            for (let i = 0; i !== raw.length; i++) {
                const lastChar = raw[i];
                if (!lastChar.match(allowed)) {
                    pendingRemoveIdx.add(i);
                }
            }
        }
        raw = raw.split('').filter((_, idx) => !pendingRemoveIdx.has(idx)).join('');
    }
    let idx = 0;
    let result = '';
    for (const char of template) {
        if (idx === raw.length) {
            break;
        }
        else if (char === symbol || char === raw[idx]) {
            result += raw[idx];
            idx++;
        }
        else {
            result += char;
        }
    }
    return result;
};

/**
 * Delays the execution for the specified amount of time.
 *
 * @param [timeout=1000] - The duration to wait in milliseconds.
 * @returns A promise that resolves once the timeout has elapsed.
 */
const sleep = (timeout = 1000) => new Promise((res) => setTimeout(() => res(), timeout));

const TIMEOUT_SYMBOL = Symbol('timeout');
const timeout = (run, delay = 30000) => {
    const wrappedFn = async (...args) => {
        const result = await Promise.race([
            run(...args),
            sleep(delay).then(() => TIMEOUT_SYMBOL),
        ]);
        return result;
    };
    return wrappedFn;
};

/**
 * Creates an awaiter object along with a promise.
 *
 * @template T - The type of the value to be resolved.
 *
 * @function
 * @returns An array containing the promise and the awaiter object.
 */
const createAwaiter = () => {
    let resolveRef = null;
    let rejectRef = null;
    const awaiter = new Promise((resolve, reject) => {
        [resolveRef, rejectRef] = [resolve, reject];
    });
    const resolve = (value) => resolveRef && resolveRef(value);
    const reject = (value) => rejectRef && rejectRef(value);
    return [awaiter, { resolve, reject }];
};

const waitForNext = async (subject, condition, delay = 0) => {
    let unsubscribeRef;
    let isFinished = false;
    const [promise, { resolve, reject }] = createAwaiter();
    unsubscribeRef = subject.subscribe((value) => {
        if (condition(value)) {
            unsubscribeRef && unsubscribeRef();
            isFinished = true;
            resolve(value);
        }
    });
    delay && sleep(delay).then(() => {
        if (isFinished) {
            return;
        }
        unsubscribeRef && unsubscribeRef();
        resolve(TIMEOUT_SYMBOL);
    });
    return promise;
};

/**
 * Returns the error message from the given error object.
 *
 * @param error - The error object.
 * @returns The error message.
 */
const getErrorMessage = (error) => {
    let errorMessage = 'Unknown error';
    if (typeof error === 'string') {
        errorMessage = error;
    }
    else if (typeof error === 'object' && error !== null) {
        if (error?.error?.message) {
            errorMessage = error.error.message;
        }
        else if (error?.data?.message) {
            errorMessage = error.data.message;
        }
        else if (error?.message) {
            errorMessage = error.message;
        }
        else {
            errorMessage = error.toString();
        }
    }
    return errorMessage;
};

/**
 * Creates a function that is only executed once, and then memoizes and returns the result.
 *
 * @template T - The type of the function to be executed once.
 * @param run - The function to be executed once.
 * @returns - The executed function with additional "clear" method to reset the execution state.
 */
const singleshot = (run) => {
    let hasRunned = false;
    let result = null;
    /**
     * Runs a function once and caches the result for subsequent calls.
     *
     * @param args - Arguments passed to the function.
     * @returns - The result of the function.
     */
    const fn = (...args) => {
        if (!hasRunned) {
            hasRunned = true;
            result = run(...args);
        }
        return result;
    };
    /**
     * Clears the given array or object by removing all its properties or elements.
     *
     * @param target - The array or object to be cleared.
     * @throws If the target is not an array or an object.
     *
     */
    fn.clear = () => {
        hasRunned = false;
    };
    return fn;
};

/**
 * A class representing a task.
 *
 * @class
 */
class Task {
    /**
     * Retrieves the current status value.
     *
     * @return The value of the status.
     */
    get status() {
        return this._status;
    }
    ;
    /**
     * Constructor for creating an instance of the class.
     *
     * @param target - The target promise to be handled.
     *
     * @return - This method does not return any value.
     */
    constructor(target) {
        this.target = target;
        this._status = "pending";
        target.then(() => this._status = "fulfilled");
        target.catch(() => this._status = "rejected");
    }
    ;
}
/**
 * Represents a higher-order function that runs a task only once and provides a way to clear the result.
 * @template T - The function type.
 * @param run - The function to be executed.
 * @returns - The wrapped function with additional clear functionality.
 */
const singlerun = (run) => {
    let result = undefined;
    /**
     * Executes the given `run` function with the provided arguments and returns the result.
     *
     * If the `result` status is not "pending", a new `Task` is created with the `run` function.
     *
     * @param args - The arguments to be passed to the `run` function.
     * @returns - The target property of the `result` object, if available.
     */
    const fn = (...args) => {
        if (result?.status !== "pending") {
            result = new Task(run(...args));
        }
        return result?.target;
    };
    /**
     * Clears the content of the given input field.
     *
     * @param input - The input field to be cleared.
     * @returns
     */
    fn.clear = () => {
        result = undefined;
    };
    /**
     * Gets the running task status
     * @returns
     */
    fn.getStatus = () => {
        return result?.status || "ready";
    };
    return fn;
};

/**
 * Symbol representing cancellation status.
 *
 * @type {Symbol}
 * @name CANCELED_PROMISE_SYMBOL
 */
const CANCELED_PROMISE_SYMBOL = Symbol('cancelable-canceled');
/**
 * Wraps a promise function and provides cancellation functionality.
 *
 * @param promise - The promise function to wrap.
 * @returns The wrapped function with cancellation capability.
 * @template T - The type of the promise's resolved value.
 * @template P - The type of the promise function's arguments.
 */
const cancelable = (promise) => {
    let cancelRef;
    /**
     * A function that wraps a given promise with cancellation functionality.
     *
     * @template P - The type of the arguments passed to the wrapped function.
     * @template T - The type of the resolved value of the promise.
     * @param promise - The promise to be wrapped.
     * @returns A promise that resolves with the resolved value of the given promise, or with the symbol `CANCELED_PROMISE_SYMBOL` if the wrapped function's
     * cancellation is triggered.
     */
    const wrappedFn = (...args) => new Promise((resolve, reject) => {
        let hasCanceled = false;
        cancelRef && cancelRef();
        cancelRef = () => {
            hasCanceled = true;
            resolve(CANCELED_PROMISE_SYMBOL);
        };
        const result = promise(...args);
        result.then((val) => {
            if (!hasCanceled) {
                resolve(val);
                return;
            }
        });
        result.catch((error) => {
            if (!hasCanceled) {
                reject(error);
                return;
            }
        });
    });
    /**
     * Cancels the execution of the wrapped function.
     *
     * @function
     * @name wrappedFn.cancel
     * @returns
     */
    wrappedFn.cancel = () => {
        cancelRef && cancelRef();
    };
    return wrappedFn;
};

const REQUEST_ANIMATION_FRAME_FN = "requestAnimationFrame" in globalThis ? globalThis.requestAnimationFrame : setTimeout;
const CANCEL_ANIMATION_FRAME_FN = "cancelAnimationFrame" in globalThis ? globalThis.cancelAnimationFrame : clearTimeout;
const REQUEST_ANIMATION_FRAME = (fn) => REQUEST_ANIMATION_FRAME_FN(fn);
const CANCEL_ANIMATION_FRAME = (id) => CANCEL_ANIMATION_FRAME_FN(id);
/**
 * Creates a debounced version of a function.
 *
 * @template T - The type of the original function.
 * @param run - The function to debounce.
 * @param [delay=1000] - The delay in milliseconds before executing the debounced function.
 * @returns - The debounced function with additional methods for clearing and flushing.
 */
const debounce = (run, delay = 1000) => {
    let timeout;
    let lastRun = null;
    const on = delay ? setTimeout : REQUEST_ANIMATION_FRAME;
    const un = delay ? clearTimeout : CANCEL_ANIMATION_FRAME;
    /**
     * Wrapper function that delays the execution of a given function
     *
     * @param run - The function to be executed
     * @param delay - The delay in milliseconds
     * @returns
     */
    const wrappedFn = (...args) => {
        timeout !== null && un(timeout);
        const exec = () => {
            lastRun = null;
            timeout = null;
            run(...args);
        };
        lastRun = exec;
        timeout = on(exec, delay);
    };
    /**
     * Clears the wrapped function from any saved state or implementation.
     *
     * @memberof wrappedFn
     */
    wrappedFn.clear = () => {
        timeout !== null && un(timeout);
        timeout = null;
        lastRun = null;
    };
    /**
     * Flushes any queued functions within the wrapped function.
     *
     * @param wrappedFn - The wrapped function that may have queued functions.
     * @returns
     */
    wrappedFn.flush = () => {
        timeout !== null && un(timeout);
        lastRun && lastRun();
        timeout = null;
        lastRun = null;
    };
    wrappedFn.pending = () => {
        return !!lastRun;
    };
    return wrappedFn;
};

/**
 * Compose multiple functions together to create a new function that applies the given functions from right to left.
 * If no functions are given, the composed function will simply return the input argument.
 * If only one function is given, the composed function will simply return the output of that function.
 *
 * @param funcs - The functions to be composed.
 * @returns - The composed function.
 */
const compose = (...funcs) => {
    if (funcs.length === 0) {
        return (arg) => arg;
    }
    if (funcs.length === 1) {
        return funcs[0];
    }
    return funcs.reduce((a, b) => (...args) => a(b(...args)));
};

/**
 * Creates a wrapper function for a Promise that allows for cancellation and clearing of queued Promises.
 *
 * @template T - The resolved value of the Promise.
 * @template P - The types of the arguments passed to the promise function.
 * @param promise - The promise function to be wrapped.
 * @returns - The wrapped function.
 */
const queued = (promise) => {
    let lastPromise = Promise.resolve();
    let cancelFn = undefined;
    /**
     * Executes a promise function with the given arguments, allowing cancellation and clearing.
     * @template P - The tuple type of the arguments passed to the promise function.
     * @param promise - The promise function to be executed.
     * @param args - The arguments to be passed to the promise function.
     * @returns - A promise that resolves with the result of the promise function or CANCELED_PROMISE_SYMBOL if canceled.
     */
    const wrappedFn = (...args) => {
        let isCanceled = false;
        const cancel = () => { isCanceled = true; };
        cancelFn = cancelFn ? compose(cancelFn, cancel) : cancel;
        lastPromise = lastPromise
            .then(async () => {
            if (!isCanceled) {
                return await promise(...args);
            }
            return CANCELED_PROMISE_SYMBOL;
        })
            .finally(() => {
            wrappedFn.clear();
            cancelFn = undefined;
        });
        return lastPromise;
    };
    /**
     * Clears any existing wrapped function stored in the wrappedFn variable.
     * This function is used to reset the wrapped function to its initial state.
     *
     * @function clear
     */
    wrappedFn.clear = () => {
        lastPromise = Promise.resolve();
        cancelFn = undefined;
    };
    /**
     * Cancels the execution of the wrapped function.
     * Once called, the wrapped function will not be executed.
     *
     * @function cancel
     * @memberof wrappedFn
     *
     */
    wrappedFn.cancel = () => {
        lastPromise = Promise.resolve();
        cancelFn && cancelFn();
        cancelFn = undefined;
    };
    return wrappedFn;
};

/**
 * Creates an execution pool for asynchronous functions with a limited concurrency.
 *
 * @template T - The type of the result of the wrapped function.
 * @template P - The types of the parameters of the wrapped function.
 *
 * @function
 * @param run - The function to be executed in the pool.
 * @param options - Optional configuration options for the execution pool.
 * @returns A wrapped function that executes asynchronously within the execution pool.
 */
const execpool = (run, { maxExec = 3, delay = 10, } = {}) => {
    const execSet = new Set();
    const execStack = [];
    /**
     * Executes a function with arguments and adds it to the execution pool.
     *
     * @function
     * @param awaiter - The awaiter to resolve the function execution.
     * @param args - The arguments to pass to the function.
     */
    const execute = async (awaiter, ...args) => {
        const exec = run(...args)
            .then((value) => {
            awaiter.resolve(value);
            execSet.delete(exec);
            return value;
        })
            .catch((reason) => {
            awaiter.reject(reason);
            execSet.delete(exec);
            return null;
        });
        execSet.add(exec);
        return await exec;
    };
    /**
     * Initializes the execution loop for the execution pool.
     *
     * @function
     */
    const initLoop = singlerun(async () => {
        while (execStack.length) {
            if (execSet.size) {
                await Promise.race(execSet);
            }
            if (execSet.size >= maxExec) {
                if (delay) {
                    await sleep(delay);
                }
                continue;
            }
            const { args, awaiter } = execStack.pop();
            await execute(awaiter, ...args);
        }
    });
    /**
     * The wrapped function that executes within the execution pool.
     *
     * @function
     * @param args - The arguments to pass to the wrapped function.
     * @returns A promise that resolves with the result of the wrapped function.
     */
    const wrappedFn = async (...args) => {
        const [result, awaiter] = createAwaiter();
        if (execSet.size < maxExec) {
            await execute(awaiter, ...args);
        }
        else {
            execStack.unshift({
                awaiter,
                args,
            });
            await initLoop();
        }
        return await result;
    };
    /**
     * Clears all pending executions in the execution pool.
     *
     * @function
     */
    wrappedFn.clear = () => {
        while (execStack.length) {
            execStack.pop();
        }
        execSet.clear();
        initLoop.clear();
    };
    return wrappedFn;
};

/**
 * Retries a function multiple times until it succeeds or reaches the maximum number of retries.
 *
 * @param run - The function to run.
 * @param count - The maximum number of retries (default is 5).
 * @returns - The wrapped function that can be canceled.
 */
const retry = (run, count = 5, delay = 1000, condition = (error) => true) => {
    const wrappedFn = queued(async (...args) => {
        let total = count;
        /**
         * Calls the function `run` repeatedly until it successfully completes or `total` reattempts have been made.
         *
         * @async
         * @returns A Promise that resolves with the result of the successful call to `run`.
         * @throws If the limit of reattempts is reached and `run` still throws an error.
         */
        const call = async () => {
            try {
                return await run(...args);
            }
            catch (error) {
                if (!condition(error)) {
                    throw error;
                }
                await sleep(delay);
                if (--total === 0) {
                    throw error;
                }
                return await call();
            }
        };
        return await call();
    });
    return wrappedFn;
};

/**
 * Caches the result of a function based on the change of arguments.
 * @template T - The type of the function to be cached.
 * @template A - The type of the arguments of the function.
 * @param changed - Function to determine if the arguments have changed.
 * @param run - The function to be cached.
 * @returns - The cached function with additional clear method.
 */
const cached = (changed, run) => {
    let lastArgs = null;
    let initial = true;
    let lastValue;
    /**
     * Clears the value of the lastArgs variable.
     * @function
     * @name clear
     * @returns
     */
    const clear = () => {
        lastArgs = null;
    };
    /**
     * Executes a function with given arguments and caches the result.
     *
     * @param args - The arguments to be passed to the function.
     * @returns - The result of the executed function.
     */
    const executeFn = (...args) => {
        if (!initial) {
            if (!changed(lastArgs, args)) {
                return lastValue;
            }
        }
        lastArgs = args;
        initial = false;
        return lastValue = run(...args);
    };
    executeFn.clear = clear;
    return executeFn;
};

/**
 * Defines the GET_VALUE_MAP constant.
 *
 * This symbol is used to uniquely identify the 'get-value-map' property in an object or map.
 * It can be used as a key to retrieve or set a value from a map.
 *
 * @const {symbol} GET_VALUE_MAP - The symbol representing the 'get-value-map' property.
 */
const GET_VALUE_MAP = Symbol('get-value-map');
/**
 * A memoization function that caches the result of a function based on its arguments.
 *
 * @template T - The function type that will be memoized
 * @template A - The argument types of the function
 * @template K - The key type used to store the memoized results
 * @param key - A function that generates a unique key based on the arguments of the original function
 * @param run - The original function to be memoized
 * @returns - A memoized version of the original function with the ability to clear the cache
 */
const memoize = (key, run) => {
    /**
     * A map that associates keys of type K with values of type IRefMemoize<ReturnType<T>>.
     *
     * @template K - The type of the keys in the valueMap.
     * @template T - The type of the values in the valueMap.
     */
    const valueMap = new Map();
    /**
     * Clears the value map.
     * If a key is provided, it deletes the corresponding value from the map.
     * If no key is provided, it clears the entire map.
     *
     * @param [key] - The key of the value to delete from the map.
     * @returns
     */
    const clear = (key) => {
        if (key) {
            valueMap.delete(key);
            return;
        }
        valueMap.clear();
    };
    /**
     * Executes a function with the given arguments and caches the result.
     * Implements the `IClearableMemoize` interface.
     *
     * @template A - The argument types of the function.
     * @template T - The return type of the function.
     * @param args - The arguments to pass to the function.
     * @returns - The cached result of the function.
     */
    const executeFn = (...args) => {
        const k = key(args);
        let value = valueMap.get(k)?.current;
        if (value === undefined) {
            const ref = { current: undefined };
            valueMap.set(k, ref);
            value = ref.current = run(...args);
        }
        return value;
    };
    /**
     * Executes the GET_VALUE_MAP function by symbol. Works like a friend classes in C++
     *
     * @return The value map containing key-value pairs.
     */
    executeFn[GET_VALUE_MAP] = () => valueMap;
    /**
     * Clears the executeFn function.
     *
     * @function clear
     * @memberof executeFn
     * @returns
     */
    executeFn.clear = clear;
    executeFn.add = (key, value) => {
        let ref = valueMap.get(key);
        if (ref === undefined) {
            ref = { current: value };
        }
        else {
            ref.current = value;
        }
        valueMap.set(key, ref);
    };
    executeFn.remove = (key) => {
        return valueMap.delete(key);
    };
    executeFn.has = (key) => {
        return valueMap.has(key);
    };
    return executeFn;
};

const CATCH_SYMBOL = Symbol("default");
/**
 * Asynchronously waits for a promise to resolve and handles any errors that occur.
 *
 * @param value - The promise to await.
 * @param config - The configuration options.
 * @param config.fallback - The fallback function to call if an error occurs.
 * @param config.defaultValue - The default value to return if an error occurs.
 * @returns - A promise that resolves to the resolved value of the input promise, or the defaultValue if an error occurs.
 *
 */
const awaiter = async (value, { fallback, defaultValue }) => {
    try {
        return await value;
    }
    catch (error) {
        fallback && fallback(error);
        return defaultValue;
    }
};
/**
 * A higher-order function that wraps the provided function with a try-catch block. It catches any errors that occur during the execution of the function and handles them according to
 * the specified configuration.
 *
 * @template T - The type of the function being wrapped
 * @template A - An array of arguments that the function accepts
 * @template V - The type of the value returned by the function
 * @template D - The type of the default value to return in case of error
 *
 * @param run - The function to be wrapped
 * @param config - The configuration object
 * @param config.fallback - The fallback function to be called with the caught error (optional)
 * @param config.defaultValue - The default value to be returned if an error occurs
 *
 * @returns - The wrapped function that handles errors and returns the result or the default value
 */
const trycatch = (run, { allowedErrors, fallback, defaultValue = CATCH_SYMBOL, } = {}) => {
    return (...args) => {
        try {
            const result = run(...args);
            if (result instanceof Promise) {
                return awaiter(result, { fallback, defaultValue });
            }
            return result;
        }
        catch (error) {
            fallback && fallback(error);
            if (allowedErrors) {
                for (const BaseError of allowedErrors) {
                    if (error instanceof BaseError) {
                        return defaultValue;
                    }
                }
                throw error;
            }
            return defaultValue;
        }
    };
};

/**
 * Represents the default timeout value in milliseconds.
 *
 * @constant {number} DEFAULT_TIMEOUT - The value of the default timeout.
 */
const DEFAULT_TIMEOUT = 250;
/**
 * A unique symbol representing a value that should never occur.
 *
 * @type {symbol}
 */
const NEVER_VALUE$7 = Symbol('never');
/**
 * Wrap a function with time-to-live (TTL) caching.
 *
 * @template T - The function type.
 * @template A - The argument types of the function.
 * @template K - The key type for caching.
 * @param run - The function to wrap.
 * @param options - The configuration options.
 * @param [options.key] - The key generator function that generates a key based on function arguments.
 * @param [options.timeout] - The TTL duration in milliseconds.
 * @returns - The wrapped function with caching capability.
 */
const ttl = (run, { key = () => NEVER_VALUE$7, timeout = DEFAULT_TIMEOUT, } = {}) => {
    /**
     * Creates a memoized function that caches the result of the
     * original function based on the provided key.
     *
     * @param key - The key used to cache the result of the function.
     * @param run - The original function to be memoized.
     * @returns - A memoized function that returns the cached value.
     */
    const wrappedFn = memoize(key, (...args) => ({
        value: run(...args),
        ttl: Date.now(),
    }));
    /**
     * Executes a wrapped function with a TTL (Time To Live).
     * @param args - The arguments for the wrapped function.
     * @returns - The return value of the wrapped function.
     */
    const executeFn = (...args) => {
        const currentTtl = Date.now();
        const { value, ttl } = wrappedFn(...args);
        if (currentTtl - ttl > timeout) {
            const k = key(args);
            wrappedFn.clear(k);
            return executeFn(...args);
        }
        return value;
    };
    /**
     * Clears the executeFn function.
     *
     * @function
     * @memberof executeFn
     * @name clear
     *
     * @returns
     */
    executeFn.clear = (key) => {
        wrappedFn.clear(key);
    };
    /**
     * Executes a garbage collection in the ttl storage.
     *
     * @function executeFn.gc
     * @returns
     */
    executeFn.gc = () => {
        const valueMap = wrappedFn[GET_VALUE_MAP]();
        for (const [key, item] of valueMap.entries()) {
            const currentTtl = Date.now();
            if (currentTtl - item.current.ttl > timeout) {
                wrappedFn.clear(key);
            }
        }
    };
    executeFn.add = (key, value) => wrappedFn.add(key, {
        value,
        ttl: Date.now(),
    });
    executeFn.remove = wrappedFn.remove;
    return executeFn;
};

const DEFAULT_DELAY = 1000;
const DEFAULT_NAME = "unknown-rate";
const NEVER_VALUE$6 = Symbol('never');
class RateError extends Error {
    constructor() {
        super(...arguments);
        this.type = "rate-error";
    }
}
const rate = (run, { key = () => NEVER_VALUE$6, rateName = DEFAULT_NAME, delay = DEFAULT_DELAY, } = {}) => {
    const wrappedFn = memoize((args) => key(args.slice(1)), (tick, ...args) => ({
        tick,
        value: run(...args),
        when: Date.now(),
    }));
    /**
     * Executes a wrapped function with a TTL (Time To Live).
     * @param args - The arguments for the wrapped function.
     * @returns - The return value of the wrapped function.
     */
    const executeFn = (...args) => {
        const currentTime = Date.now();
        const currentTick = randomString();
        const { value, when, tick } = wrappedFn(currentTick, ...args);
        if (tick === currentTick) {
            return value;
        }
        if (currentTime - when < delay) {
            throw new RateError(`functools-kit rate rateName=${rateName} delay not reached`);
        }
        const k = key(args);
        wrappedFn.clear(k);
        return executeFn(...args);
    };
    /**
     * Clears the executeFn function.
     *
     * @function
     * @memberof executeFn
     * @name clear
     *
     * @returns
     */
    executeFn.clear = (key) => {
        wrappedFn.clear(key);
    };
    /**
     * Executes a garbage collection in the ttl storage.
     *
     * @function executeFn.gc
     * @returns
     */
    executeFn.gc = () => {
        const valueMap = wrappedFn[GET_VALUE_MAP]();
        for (const [key, item] of valueMap.entries()) {
            const currentRate = Date.now();
            if (currentRate - item.current.ttl > delay) {
                wrappedFn.clear(key);
            }
        }
    };
    executeFn.add = (key, value) => wrappedFn.add(key, {
        value,
        tick: randomString(),
        when: Date.now(),
    });
    executeFn.remove = wrappedFn.remove;
    return executeFn;
};

/**
 * Throttle function execution to a specific delay.
 * @template T - Function type
 * @param run - Function to be throttled
 * @param delay - Delay in milliseconds (default: 1000)
 * @returns - Throttled function with clear method
 */
const throttle = (run, delay = 1000) => {
    let lastExec = 0;
    /**
     * Clears the existing timeout.
     *
     * @function
     * @name clearExistingTimeout
     * @returns
     */
    const clearExistingTimeout = () => {
    };
    /**
     * A wrapper function that delays the execution of the given function
     * until a certain amount of time has passed since the last execution.
     *
     * @param args - Arguments to be passed to the wrapped function.
     * @returns
     */
    const wrappedFn = (...args) => {
        let elapsed = Date.now() - lastExec;
        const exec = () => {
            lastExec = Date.now();
            run(...args);
        };
        {
            exec();
        }
        if (elapsed > delay) {
            exec();
        }
    };
    /**
     * Clears the wrapped function.
     * This function removes any existing functionality from the wrapped function.
     *
     * @memberof wrappedFn
     * @function clear
     */
    wrappedFn.clear = clearExistingTimeout;
    return wrappedFn;
};

const COMMIT_CHECK_DELAY = 1000;
const schedule = (run, { onSchedule, delay = COMMIT_CHECK_DELAY }) => {
    const executeFn = singlerun(run);
    let argsLast = null;
    const commitQueue = queued(async (args) => {
        if (argsLast) {
            await onSchedule(argsLast);
        }
        argsLast = args;
    });
    const waitForSchedule = singlerun(async () => {
        while (true) {
            if (executeFn.getStatus() !== "pending") {
                break;
            }
            if (!argsLast) {
                return null;
            }
            await sleep(delay);
        }
        const args = argsLast;
        argsLast = null;
        return await executeFn(...args);
    });
    const wrappedFn = async (...args) => {
        if (executeFn.getStatus() === "pending") {
            await commitQueue(args);
            return await waitForSchedule();
        }
        return await executeFn(...args);
    };
    wrappedFn.clear = () => {
        argsLast = null;
    };
    return wrappedFn;
};

const PUBSUB_TIMEOUT = 30000;
class PubsubArrayAdapter {
    constructor(_maxItems = Number.POSITIVE_INFINITY) {
        this._maxItems = _maxItems;
        this._array = [];
        this.length = () => Promise.resolve(this._array.length);
        this.push = async (value) => {
            this._array.push(value);
            while (await this.length() > this._maxItems) {
                this._array.shift();
            }
            return Promise.resolve();
        };
        this.pop = async () => {
            const value = this._array.pop();
            return Promise.resolve(value);
        };
        this.shift = () => Promise.resolve(this._array.shift());
        this.clear = () => {
            while (this._array.length) {
                this._array.pop();
            }
            return Promise.resolve();
        };
    }
    getFirst() {
        const [first] = this._array;
        return Promise.resolve(first || null);
    }
    ;
    async *[Symbol.asyncIterator]() {
        for (const item of this._array) {
            yield item;
        }
    }
    ;
}
class PubsubMapAdapter {
    constructor() {
        this._map = new Map();
    }
    clear() {
        this._map.clear();
        return Promise.resolve();
    }
    getFirst() {
        const iterator = this._map.entries();
        const first = iterator.next();
        return Promise.resolve(first.done ? null : first.value);
    }
    size() {
        return Promise.resolve(this._map.size);
    }
    set(key, value) {
        this._map.set(key, value);
        return Promise.resolve();
    }
    get(key) {
        const value = this._map.get(key) ?? null;
        return Promise.resolve(value);
    }
    delete(key) {
        this._map.delete(key);
        return Promise.resolve();
    }
    shift() {
        const iterator = this._map.entries();
        const first = iterator.next();
        if (first.done) {
            return Promise.resolve(null);
        }
        this._map.delete(first.value[0]);
        return Promise.resolve(first.value);
    }
    async *[Symbol.asyncIterator]() {
        for (const entry of this._map.entries()) {
            yield entry;
        }
    }
}
const pubsub = (emitter, { onDestroy, onBegin, onProcess, onEnd, timeout = PUBSUB_TIMEOUT, queue: initialQueue = new PubsubArrayAdapter(), } = {}) => {
    const awaiterMap = new Map();
    const queue = initialQueue;
    let lastOk = Date.now();
    let isStopped = false;
    const handleStop = async () => {
        if (isStopped) {
            return;
        }
        isStopped = true;
        if (onDestroy) {
            await onDestroy(queue);
        }
        await queue.clear();
        awaiterMap.clear();
    };
    const makeBroadcast = singlerun(async () => {
        if (isStopped) {
            return;
        }
        while (await queue.length()) {
            const first = await queue.getFirst();
            if (!first) {
                break;
            }
            const [id, data] = first;
            const awaiter = awaiterMap.get(id);
            if (!awaiter) {
                console.error("functools-kit pubsub missing awaiter", { id, data });
                continue;
            }
            let success = false;
            try {
                success = await emitter(data);
            }
            catch {
                success = false;
            }
            finally {
                await sleep(10);
            }
            if (success && onEnd) {
                await onEnd(data);
            }
            if (success) {
                lastOk = Date.now();
                await queue.shift();
                awaiterMap.delete(id);
                awaiter.resolve();
                continue;
            }
            await sleep(1000);
            if (Date.now() - lastOk >= timeout) {
                await handleStop();
                return;
            }
        }
    });
    const makeCommit = async (data) => {
        if (onBegin) {
            await onBegin(data);
        }
        const [result, awaiter] = createAwaiter();
        const id = randomString();
        awaiterMap.set(id, awaiter);
        await queue.push([id, data]);
        if (onProcess) {
            await onProcess(data);
        }
        await makeBroadcast();
        return await result;
    };
    const makeInit = singleshot(async () => {
        const resolveList = [];
        for await (const [id] of queue) {
            const [resolve, awaiter] = createAwaiter();
            awaiterMap.set(id, awaiter);
            resolveList.push(resolve);
        }
        await makeBroadcast();
        await Promise.all(resolveList);
    });
    const wrappedFn = async (data) => {
        if (isStopped) {
            return;
        }
        await makeInit();
        await makeCommit(data);
    };
    wrappedFn.stop = handleStop;
    return wrappedFn;
};
pubsub.fromMap = (map) => new class {
    constructor() {
        this.clear = async () => await map.clear();
        this.getFirst = async () => await map.getFirst();
        this.length = async () => await map.size();
        this.push = async ([key, value]) => await map.set(key, value);
        this.shift = async () => await map.shift();
    }
    async *[Symbol.asyncIterator]() {
        for await (const entry of map) {
            yield entry;
        }
    }
};

const RANDOM_CODE = randomString();
const OBSOLETE_PROMPT = `Method marked as obsolete. Type ${RANDOM_CODE} to continue`;
const getMethodName = (fn, self) => {
    if (!self) {
        return "unknown";
    }
    if (typeof self !== 'object') {
        return "unknown";
    }
    for (const key of Object.keys(self)) {
        if (self[key] === fn) {
            return key;
        }
    }
    return getMethodName(fn, Object.getPrototypeOf(self));
};
const obsolete = (run, methodName = "") => {
    const fn = async (...args) => {
        methodName = methodName || getMethodName(fn, undefined);
        console.log(`react-declarative obsolete hof call method_name=${methodName}`);
        if ('prompt' in globalThis) {
            if (globalThis.prompt(OBSOLETE_PROMPT) === RANDOM_CODE) {
                return await run(...args);
            }
        }
        throw new Error('react-declarative obsolete hof call');
    };
    return fn;
};

/**
 * Wraps a function with a single event loop ticking behavior.
 *
 * @param run - The function to be wrapped.
 * @returns A wrapped function that executes with a single ticking behavior.
 */
const singletick = (run) => {
    let timeout = null;
    const singleshotFn = singleshot(run);
    /**
     * This is a higher-order function that wraps another function and adds
     * a delay before invoking it.
     *
     * @param singleshotFn - The function to be wrapped.
     * @param singletick.delay - The delay in milliseconds before invoking the wrapped function.
     * @returns - The result of invoking the wrapped function.
     */
    const wrappedFn = (...args) => {
        const result = singleshotFn(...args);
        timeout !== null && clearTimeout(timeout);
        timeout = setTimeout(() => {
            singleshotFn.clear();
            timeout = null;
        }, singletick.delay);
        return result;
    };
    return wrappedFn;
};
singletick.delay = 0;

/**
 * Creates a wrapped function that only executes the provided function after the initial call has completed.
 * The wrapped function can be cleared to allow subsequent calls to execute the provided function again.
 *
 * @template T The type of the promise resolved by the provided function.
 * @template P The type of the arguments passed to the provided function.
 * @param run The function to be wrapped.
 * @returns The wrapped function.
 */
const afterinit = (run) => {
    let hasComplete = false;
    /**
     * A wrapper function that either resolves a Promise or calls the inner function 'run' depending on its state.
     *
     * @param args - The arguments to be passed to the inner function 'run'.
     * @returns - A Promise that resolves to 'undefined' if 'hasComplete' is false, otherwise calls the inner function 'run'.
     */
    const wrappedFn = (...args) => {
        if (!hasComplete) {
            hasComplete = true;
            return Promise.resolve();
        }
        return run(...args);
    };
    /**
     * Clears the function wrapped in a wrapper function.
     * This function removes any previously set function and resets it to null.
     *
     * @param wrappedFn - The wrapped function to be cleared.
     * @returns
     */
    wrappedFn.clear = () => {
        hasComplete = false;
    };
    return wrappedFn;
};

/**
 * Returns the first element of an array.
 *
 * @template T - The type of array elements.
 * @param arr - The input array.
 * @returns - The first element of the array, or null if the array is null or empty.
 */
const first = (arr) => {
    if (Array.isArray(arr)) {
        const [first] = arr;
        return first || null;
    }
    return null;
};

/**
 * Checks if an object is empty.
 *
 * @param obj - The object to check for emptiness.
 * @returns - Returns true if the object is empty, otherwise returns false.
 */
function isEmpty(obj) {
    if (Object.getOwnPropertySymbols(obj).length !== 0) {
        return false;
    }
    if (Object.getOwnPropertyNames(obj).length !== 0) {
        return false;
    }
    return true;
}

/**
 * Class representing an event emitter.
 * @class
 */
class EventEmitter {
    constructor() {
        this._events = {};
    }
    /**
     * Check if the object has any listeners attached to it.
     *
     * @return True if the object has listeners, false otherwise.
     */
    get hasListeners() {
        return !isEmpty(this._events);
    }
    ;
    /**
     * Retrieves the listeners associated with the given event key.
     *
     * @param key - The event key to retrieve the listeners for.
     * @returns An array of listeners associated with the given event key.
     */
    getListeners(key) {
        return this._events[key] || [];
    }
    ;
    /**
     * Subscribes a callback function to the specified event name.
     *
     * @param eventName - The key of the event.
     * @param callback - The callback function to be executed when the event is triggered.
     * @returns
     */
    subscribe(eventName, callback) {
        !this._events[eventName] && (this._events[eventName] = []);
        this._events[eventName].push(callback);
    }
    ;
    /**
     * Removes a callback function from the list of event listeners for the specified event.
     *
     * @param eventName - The key of the event to unsubscribe from.
     * @param callback - The callback function to remove from the event listeners.
     * @returns
     */
    unsubscribe(eventName, callback) {
        !this._events[eventName] && (this._events[eventName] = []);
        this._events[eventName] = this._events[eventName].filter(eventCallback => callback !== eventCallback);
    }
    ;
    /**
     * Clears all event handlers registered for the current object.
     * @function
     * @returns
     */
    unsubscribeAll() {
        this._events = {};
    }
    ;
    /**
     * Subscribes a callback function to the given event name. The callback function will be triggered only once when the event is emitted.
     *
     * @param eventName - The name of the event to subscribe to.
     * @param callback - The callback function to be executed when the event is emitted.
     * @returns - A function that can be called to unsubscribe the callback function from the event.
     */
    once(eventName, callback) {
        const subscriber = async (...args) => {
            await callback(...args);
            this.unsubscribe(eventName, subscriber);
        };
        this.subscribe(eventName, subscriber);
        return () => {
            this.unsubscribe(eventName, subscriber);
        };
    }
    ;
    /**
     * Emits the given event with the specified arguments.
     *
     * @param eventName - The name of the event to emit.
     * @param args - The arguments to pass to the event listeners.
     * @returns - A promise that resolves when all event listeners have completed.
     */
    async emit(eventName, ...args) {
        const event = [...this._events && this._events[eventName] || []];
        for (let i = 0; i !== event.length; i++) {
            await event[i](...args);
        }
    }
    ;
}

const OBSERVER_EVENT = Symbol('observer-subscribe');
const CONNECT_EVENT = Symbol('observer-connect');
const DISCONNECT_EVENT = Symbol('observer-disconnect');
const LISTEN_CONNECT = Symbol('observer-connect-listen');
const LISTEN_DISCONNECT = Symbol('observer-disconnect-listen');
/**
 * A class representing an Observer.
 *
 * @template Data - The type of data to observe.
 */
class Observer {
    /**
     * Returns the current value of the 'isShared' property.
     *
     * @returns - The value of the 'isShared' property.
     */
    get isShared() {
        return this._isShared;
    }
    ;
    /**
     * Returns whether the given event has any listeners.
     *
     * @returns True if there are listeners for the event, otherwise false.
     */
    get hasListeners() {
        return !!this.broadcast.getListeners(OBSERVER_EVENT).length;
    }
    ;
    constructor(dispose) {
        this.dispose = dispose;
        this.broadcast = new EventEmitter();
        this._isShared = false;
        /**
         * Subscribe a given observer to the global broadcast event.
         *
         * @param observer - The observer subscribing to the event.
         * @param callback - The callback function to be executed when the event is triggered.
         * @returns
         */
        this._subscribe = (observer, callback) => {
            this.broadcast.subscribe(OBSERVER_EVENT, callback);
            observer[LISTEN_CONNECT](() => {
                this.broadcast.emit(CONNECT_EVENT);
            });
            observer[LISTEN_DISCONNECT](() => {
                if (!this.hasListeners) {
                    this.broadcast.emit(DISCONNECT_EVENT);
                }
            });
        };
        /**
         * Unsubscribes a callback function from the observer event.
         *
         * @param callback - The callback function to unsubscribe.
         * @returns
         */
        this._unsubscribe = (callback) => {
            this.broadcast.unsubscribe(OBSERVER_EVENT, callback);
        };
        /**
         * Tries to dispose the object if it has no listeners and is not shared.
         * If disposed successfully, emits the DISCONNECT_EVENT.
         */
        this.tryDispose = () => {
            if (!this.hasListeners && !this._isShared) {
                this.dispose();
                this.broadcast.emit(DISCONNECT_EVENT);
            }
        };
        /**
         * Creates a new Observer.
         * @template T - The type of the value emitted by the observer.
         * @param callbackfn - A function to apply to each value emitted by the observer.
         * @returns - The created Observer.
         */
        this.map = (callbackfn) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const handler = (value) => {
                const pendingValue = callbackfn(value);
                observer.emit(pendingValue);
            };
            this._subscribe(observer, handler);
            unsubscribeRef = () => this._unsubscribe(handler);
            return observer;
        };
        /**
         * Applies a callback function to each value emitted by the Observable and flattens the resulting values into a new Observable.
         *
         * @template T - The type of values emitted by the Observable.
         *
         * @param callbackfn - A callback function that accepts a value emitted by the Observable and returns an array of values or a single value.
         *
         * @returns - A new Observer that emits the flattened values.
         */
        this.flatMap = (callbackfn) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const handler = (value) => {
                const pendingValue = callbackfn(value);
                if (Array.isArray(pendingValue)) {
                    pendingValue.forEach((value) => {
                        observer.emit(value);
                    });
                }
                else {
                    observer.emit(pendingValue);
                }
            };
            this._subscribe(observer, handler);
            unsubscribeRef = () => this._unsubscribe(handler);
            return observer;
        };
        /**
         * Operator function to create a new observer with a transformed data type.
         *
         * @template T - The type of the transformed data.
         * @param callbackfn - A callback function that takes the target observer and returns a new observer with transformed data.
         * @returns - A new observer with the transformed data type.
         */
        this.operator = (callbackfn) => {
            return callbackfn(this);
        };
        /**
         * Reduces the data emitted by an Observer using a callback function and an initial value.
         *
         * @template T - The type of the accumulator and the return value.
         * @param callbackfn - The callback function to execute on each emitted value.
         *   It takes an accumulator value and the current value being emitted, and returns the new accumulator value.
         * @param begin - The initial value of the accumulator.
         * @returns - An Observer that emits the accumulated value after each emission.
         */
        this.reduce = (callbackfn, begin) => {
            let unsubscribeRef;
            let acm = begin;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const handler = (value) => {
                const pendingValue = callbackfn(acm, value);
                acm = pendingValue;
                observer.emit(pendingValue);
            };
            this._subscribe(observer, handler);
            unsubscribeRef = () => this._unsubscribe(handler);
            return observer;
        };
        /**
         * Creates and returns an observer function that splits an array of data
         * into a nested array of a specified length.
         *
         * @returns The split observer function.
         */
        this.split = () => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const handler = (data) => {
                if (Array.isArray(data)) {
                    data.flat(Number.POSITIVE_INFINITY).forEach((item) => {
                        observer.emit(item);
                    });
                }
                else {
                    observer.emit(data);
                }
            };
            this._subscribe(observer, handler);
            unsubscribeRef = () => this._unsubscribe(handler);
            return observer;
        };
        /**
         * Creates an Observer with asynchronous mapping functionality.
         *
         * @template T - The type of the result of the mapping function.
         * @param callbackfn - The function used to map the incoming data.
         * @param [fallbackfn] - An optional fallback function to handle error cases. If not provided, the error will be rethrown.
         * @returns - The created Observer.
         */
        this.mapAsync = (callbackfn, fallbackfn) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const iteraction = queued(callbackfn);
            const handler = async (value) => {
                try {
                    const pendingValue = await iteraction(value);
                    if (pendingValue !== CANCELED_PROMISE_SYMBOL) {
                        observer.emit(pendingValue);
                    }
                }
                catch (e) {
                    if (fallbackfn) {
                        fallbackfn(e);
                    }
                    else {
                        throw e;
                    }
                }
            };
            this._subscribe(observer, handler);
            unsubscribeRef = compose(() => this._unsubscribe(handler), () => iteraction.clear());
            return observer;
        };
        /**
         * Creates a filtered observer.
         *
         * @param callbackfn - The filter callback function.
         * @returns The filtered observer.
         */
        this.filter = (callbackfn) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const handler = (value) => {
                const delegate = callbackfn(value);
                if (delegate) {
                    observer.emit(value);
                }
            };
            this._subscribe(observer, handler);
            unsubscribeRef = () => this._unsubscribe(handler);
            return observer;
        };
        /**
         * Attaches a callback function to the tap observer. The callback function will be called with a value of type `Data` when the tap observer is triggered.
         *
         * @param callbackfn - A callback function that takes a value of type `Data` as an argument.
         * @returns - An observer object that can be used to manage the tap subscription.
         */
        this.tap = (callbackfn) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const handler = (value) => {
                callbackfn(value);
                observer.emit(value);
            };
            this._subscribe(observer, handler);
            unsubscribeRef = () => this._unsubscribe(handler);
            return observer;
        };
        /**
         * Creates a debounced observer that emits values at a specified delay.
         *
         * @param delay - The delay (in milliseconds) between value emissions.
         * @returns The debounced observer.
         */
        this.debounce = (delay) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            const handler = debounce((value) => {
                observer.emit(value);
            }, delay);
            this._subscribe(observer, handler);
            unsubscribeRef = compose(() => handler.clear(), () => this._unsubscribe(handler));
            return observer;
        };
        /**
         * Creates a delayed observer that emits values at a specified delay.
         *
         * @param delay - The delay (in milliseconds) between value emissions.
         * @returns The debounced observer.
         */
        this.delay = (delay) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const observer = new Observer(dispose);
            let isCanceled = false;
            const handler = queued(async (value) => {
                await sleep(delay);
                if (!isCanceled) {
                    await observer.emit(value);
                }
            });
            this._subscribe(observer, handler);
            unsubscribeRef = compose(() => handler.clear(), () => this._unsubscribe(handler), () => { isCanceled = true; });
            return observer;
        };
        /**
         * Emits the specified data to all observers.
         *
         * @param data - The data to be emitted.
         */
        this.emit = async (data) => {
            await this.broadcast.emit(OBSERVER_EVENT, data);
        };
        /**
         * Subscribes to the `OBSERVER_EVENT` and invokes the provided callback function.
         * Emits the `CONNECT_EVENT`.
         * Returns a composed function that will try to dispose and unsubscribe the callback.
         *
         * @param callbackfn - The callback function to be invoked when `OBSERVER_EVENT` is emitted.
         * @returns - The composed function that will try to dispose and unsubscribe the callback.
         */
        this.connect = (callbackfn) => {
            this.broadcast.subscribe(OBSERVER_EVENT, callbackfn);
            this.broadcast.emit(CONNECT_EVENT);
            return compose(() => this.tryDispose(), () => this._unsubscribe(callbackfn));
        };
        /**
         * Executes a callback function once and provides a way to unsubscribe from further executions.
         *
         * @param callbackfn - The callback function to be executed once.
         * @returns - A function that can be called to unsubscribe from further executions of the callback.
         */
        this.once = (callbackfn) => {
            let unsubscribeRef;
            const handler = (value) => {
                callbackfn(value);
                unsubscribeRef();
            };
            unsubscribeRef = this.connect(handler);
            return unsubscribeRef;
        };
        /**
         * Marks a variable as shared.
         *
         * @returns The shared variable object.
         */
        this.share = () => {
            this._isShared = true;
            return this;
        };
        /**
         * Creates an observable sequence that emits values at specified intervals.
         * @param [interval=1000] - The time interval between emissions in milliseconds.
         * @returns The observer object to subscribe to.
         */
        this.repeat = (interval = 1000) => {
            let unsubscribeRef;
            let timeout;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef(), () => timeout !== undefined && clearTimeout(timeout));
            const observer = new Observer(dispose);
            const handler = (value) => {
                if (timeout !== undefined) {
                    clearTimeout(timeout);
                }
                observer.emit(value);
                if (this.hasListeners) {
                    timeout = setTimeout(handler, interval, value);
                }
            };
            this._subscribe(observer, handler);
            unsubscribeRef = () => this._unsubscribe(handler);
            return observer;
        };
        /**
         * Merges an observer with the given observer, returning a new observer that emits values from both observers.
         *
         * @template T - The type of value emitted by the observer.
         * @param observer - The observer to merge with.
         * @returns - The merged observer.
         */
        this.merge = (observer) => {
            let unsubscribeRef;
            const dispose = compose(() => this.tryDispose(), () => unsubscribeRef());
            const merged = new Observer(dispose);
            const handler = (value) => {
                merged.emit(value);
            };
            this._subscribe(merged, handler);
            let unsubscribe = () => undefined;
            merged[LISTEN_CONNECT](() => {
                unsubscribe = observer.connect(handler) || (() => undefined);
            });
            unsubscribeRef = compose(() => this._unsubscribe(handler), () => unsubscribe());
            return merged;
        };
        /**
         * Unsubscribes from all events and performs cleanup.
         *
         * @function
         * @name unsubscribe
         * @memberOf undefined
         *
         * @returns
         */
        this.unsubscribe = () => {
            this.broadcast.unsubscribeAll();
            this.broadcast.emit(DISCONNECT_EVENT);
            this.dispose();
        };
        /**
         * Converts the current instance to a Promise that resolves with the data.
         *
         * @returns A Promise that resolves with the data.
         */
        this.toPromise = singlerun(() => new Promise((res) => {
            this.once((data) => res(data));
        }));
        /**
         * Creates a context for iterating asynchronously using a generator function.
         *
         * @returns The iterator context object.
         * @property iterate - The generator function that can be used to iterate over the values.
         * @property done - Marks the iteration as complete.
         */
        this.toIteratorContext = () => {
            const self = this;
            let isDone = false;
            const iterate = async function* () {
                while (!isDone) {
                    const next = await self.toPromise();
                    yield next;
                }
            };
            return {
                iterate,
                done() {
                    isDone = true;
                },
            };
        };
    }
    /**
     * Sets up a listener for the connect event on the broadcast channel.
     *
     * @param fn - The callback function to be executed once the connect event is triggered.
     * @returns
     */
    [LISTEN_CONNECT](fn) {
        this.broadcast.once(CONNECT_EVENT, fn);
    }
    ;
    /**
     * Adds a listener for the DISCONNECT_EVENT.
     *
     * @param fn - The function to be executed when the event occurs.
     */
    [LISTEN_DISCONNECT](fn) {
        this.broadcast.once(DISCONNECT_EVENT, fn);
    }
    ;
}

const SUBJECT_EVENT = Symbol('rn-declarative-subject');
/**
 * Represents a subject that can emit data and be subscribed to.
 * @class
 * @implements {TSubject<Data>}
 * @implements {TObservable<Data>}
 * @template Data - The type of data that the subject emits.
 */
class Subject {
    constructor() {
        this._emitter = new EventEmitter();
        /**
         * Converts an observer-based asynchronous operation into a promise-based asynchronous operation.
         *
         * @function toPromise
         * @instance
         * @returns A promise representing the completion or failure of the asynchronous operation.
         */
        this.toPromise = singlerun(() => this.toObserver().toPromise());
        this.next = this.next.bind(this);
        this.toObserver = this.toObserver.bind(this);
        this.toIteratorContext = this.toIteratorContext.bind(this);
    }
    ;
    get hasListeners() {
        return this._emitter.hasListeners;
    }
    ;
    async waitForListener() {
        while (true) {
            if (this.hasListeners) {
                break;
            }
            await sleep(500);
        }
    }
    ;
    /**
     * Maps the values of the observer using the given callback function.
     *
     * @template T - The type of the mapped values.
     * @param callbackfn - A function that maps each value of the observer.
     * @returns - An observer with the mapped values.
     */
    map(callbackfn) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.map(callbackfn);
    }
    ;
    /**
     * Applies a transformation function to each value emitted by the observer and flattens the result into a single observer.
     * @template T - The type of values emitted by the observer.
     * @param callbackfn - The transformation function to apply to each value emitted by the observer.
     * @returns - The observer that emits the flattened values.
     */
    flatMap(callbackfn) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.flatMap(callbackfn);
    }
    ;
    /**
     * Applies a reducer function to each value emitted by the observer and returns a single accumulated value.
     *
     * @template T - The type of the accumulated value and emitted values
     * @param callbackfn - A function that accepts the accumulated value and the current emitted value, and returns the new accumulated value
     * @param begin - The initial value for the accumulator
     * @returns - An observer that emits the accumulated value when the original observer completes
     */
    reduce(callbackfn, begin) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.reduce(callbackfn, begin);
    }
    ;
    /**
     * Asynchronously maps the emitted values of the observer using the provided callback function.
     *
     * @template T - The type of the mapped values.
     * @param callbackfn - The callback function that maps the emitted values of the observer.
     * @param [fallbackfn] - The optional fallback function that handles errors during mapping.
     * @returns - Returns a new observer that emits the mapped values.
     */
    mapAsync(callbackfn, fallbackfn) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.mapAsync(callbackfn, fallbackfn);
    }
    ;
    /**
     * Applies a filtering function to the observer and returns a new observer with filtered values.
     *
     * @param callbackfn - A function that tests each value in the observer. Should return true or false.
     * @returns - A new observer with filtered values.
     */
    filter(callbackfn) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.filter(callbackfn);
    }
    ;
    /**
     * The tap function allows you to perform side effects without modifying the observed data.
     *
     */
    tap(callbackfn) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.tap(callbackfn);
    }
    ;
    /**
     * Applies a callback function to the values emitted by an observer.
     *
     * @param callbackfn - The callback function to apply to the emitted values.
     * @returns - An observer with the applied operator.
     *
     * @template T - The type of values emitted by the observer.
     *
     * @category Observables
     */
    operator(callbackfn) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.operator(callbackfn);
    }
    ;
    /**
     * Splits the observed data into batches of arrays.
     *
     * @returns - The observer that emits batches of arrays.
     */
    split() {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.split();
    }
    ;
    /**
     * Creates a debounced observer with an optional delay.
     * @param [delay] - The delay in milliseconds before emitting the data.
     * @returns - The debounced observer.
     */
    debounce(delay) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.debounce(delay);
    }
    ;
    /**
     * Creates a delayed observer with an optional delay.
     * @param [delay] - The delay in milliseconds before emitting the data.
     * @returns - The delayed observer.
     */
    delay(delay) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.delay(delay);
    }
    ;
    /**
     * Creates an observer that repeats emitting values at a specified interval.
     *
     * @param [interval] - The time interval at which to repeat emitting values.
     * @returns - The created observer.
     */
    repeat(interval) {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer.repeat(interval);
    }
    ;
    /**
     * Merges the provided observer with the current observer instance.
     * Returns a new observer that emits values from both observers.
     *
     * @param observer - The observer to merge with the current observer.
     * @returns - A new observer that emits values from both observers.
     */
    merge(observer) {
        let unsubscribeRef;
        const merged = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(merged.emit);
        return merged.merge(observer);
    }
    ;
    /**
     * Subscribes to an event.
     *
     * @param callback - The callback function to be invoked when the event is triggered.
     * @returns - A function to unsubscribe from the event.
     */
    subscribe(callback) {
        this._emitter.subscribe(SUBJECT_EVENT, callback);
        return () => {
            this._emitter.unsubscribe(SUBJECT_EVENT, callback);
        };
    }
    ;
    /**
     * Unsubscribes all event listeners.
     *
     * @function unsubscribeAll
     * @instance
     * @returns - No return value.
     */
    unsubscribeAll() {
        this._emitter.unsubscribeAll();
    }
    ;
    /**
     * Executes the provided callback function only once.
     * The callback function will be invoked when the specified event occurs for the first time.
     *
     * @param callback - The function to be executed only once.
     * @returns - A function that removes the registered event listener.
     */
    once(callback) {
        return this._emitter.once(SUBJECT_EVENT, callback);
    }
    ;
    /**
     * Calls the next method to emit the specified data using the SUBJECT_EVENT event.
     *
     * @param data - The data to be emitted.
     * @return - Resolves when the emission is complete.
     */
    async next(data) {
        await this._emitter.emit(SUBJECT_EVENT, data);
    }
    ;
    /**
     * Creates a new observer to observe the data emitted by a source.
     *
     * @template TObserver - The type of observer.
     * @template Data - The type of data emitted by the source.
     * @returns - The created observer.
     */
    toObserver() {
        let unsubscribeRef;
        const observer = new Observer(() => unsubscribeRef());
        unsubscribeRef = this.subscribe(observer.emit);
        return observer;
    }
    ;
    /**
     * Converts the current object to an iterator context.
     *
     * @function
     * @returns The iterator context representing the current object.
     */
    toIteratorContext() {
        return this.toObserver().toIteratorContext();
    }
}

/**
 * Represents a BehaviorSubject that extends the Subject class and provides the functionality of an observable and an observer.
 *
 * @template Data - The type of the data that the BehaviorSubject holds.
 */
class BehaviorSubject extends Subject {
    constructor(_data = null) {
        super();
        this._data = _data;
        /**
         * Sets the given data and calls the next method of the super class asynchronously.
         *
         * @param data - The data to be set.
         * @return Resolves when super class's next method is called.
         */
        this.next = async (data) => {
            this._data = data;
            await super.next(data);
        };
        /**
         * Creates a new observer.
         *
         * @returns The observer instance.
         */
        this.toObserver = () => {
            let unsubscribeRef;
            const observer = new Observer(() => unsubscribeRef());
            observer[LISTEN_CONNECT](() => {
                this._data && observer.emit(this._data);
            });
            unsubscribeRef = this.subscribe(observer.emit);
            return observer;
        };
    }
    ;
    /**
     * Retrieves the data stored in the instance.
     *
     * @return The data stored in the instance.
     */
    get data() {
        return this._data;
    }
    ;
}

/**
 * A special symbol representing a value that should never be used or returned.
 *
 * @type {symbol}
 */
const NEVER_VALUE$5 = Symbol('never');
/**
 * Wraps a promise function with lock functionality.
 *
 * @param promise - The promise function to be wrapped.
 * @returns The wrapped function with lock functionality.
 */
const lock = (promise) => {
    let lockCount = 0;
    let lockSubject = new BehaviorSubject(lockCount);
    /**
     * Waits for the lock to be unlocked.
     * @returns - Resolves when the lock is unlocked.
     */
    const waitForUnlock = () => new Promise((resolve) => {
        const handler = () => {
            if (lockSubject.data === 0) {
                resolve();
                return;
            }
            lockSubject.once(handler);
        };
        handler();
    });
    /**
     * Executes a function, after waiting for the unlock signal.
     *
     * @function
     * @param args - The arguments to pass to the function.
     * @returns - A promise that resolves to the value returned by the function.
     */
    const executeFn = queued(async (...args) => {
        await waitForUnlock();
        if (first(args) === NEVER_VALUE$5) {
            return null;
        }
        return await promise(...args);
    });
    /**
     * Asynchronous function wrapper.
     * @param args - Arguments to be passed to the wrapped function.
     * @returns - A promise that resolves to the result of the wrapped function.
     */
    const wrappedFn = async (...args) => {
        return await executeFn(...args);
    };
    /**
     * Removes all the wrapped functions that are stored in the wrappedFn collection.
     * @function wrappedFn.clear
     * @returns
     */
    wrappedFn.clear = () => {
        lockSubject.next(0);
        lockCount = 0;
        lockSubject = new BehaviorSubject(lockCount);
        executeFn.clear();
    };
    /**
     * Cancels the execution of a wrapped function.
     *
     * @function cancel
     * @memberof wrappedFn
     *
     */
    wrappedFn.cancel = () => {
        wrappedFn.clear();
        executeFn.cancel();
    };
    /**
     * Begins a lock for the wrapped function.
     *
     * @function beginLock
     * @memberof wrappedFn
     */
    wrappedFn.beginLock = () => {
        lockCount += 1;
        lockSubject.next(lockCount);
    };
    /**
     * Releases the lock after the wrapped function has completed execution.
     *
     * @param wrappedFn - The function that is wrapped and needs the lock released.
     * @returns - The result of executing the wrapped function.
     */
    wrappedFn.endLock = async () => {
        lockCount = Math.max(lockCount - 1, 0);
        lockSubject.next(lockCount);
        await executeFn(NEVER_VALUE$5);
    };
    return wrappedFn;
};

/**
 * A utility function to make a class extendable.
 *
 * @template T - The type of the constructor.
 * @param {T} BaseClass - The base class to be extended.
 * @returns {T} - A new class that extends the base class.
 */
const makeExtendable = (BaseClass) => {
    /**
     * A safe class that extends the base class.
     *
     * @param {...ConstructorParameters<T>} args - The arguments to pass to the base class constructor.
     * @returns {InstanceType<T>} - An instance of the base class.
     */
    function SafeClass(...args) {
        const instance = Reflect.construct(BaseClass, args, new.target || SafeClass);
        return instance;
    }
    Object.setPrototypeOf(SafeClass, BaseClass);
    Object.setPrototypeOf(SafeClass.prototype, BaseClass.prototype);
    return SafeClass;
};

/**
 * A utility function to deep flatten an array of objects.
 *
 * @param arr - The input array to be deep flattened.
 * @returns - The deep flattened array.
 * @template T - The type of elements in the input array.
 */
const deepFlat = (arr = []) => {
    const result = [];
    const process = (entries = []) => entries?.forEach((entry) => {
        const fields = entry['fields'] || [];
        const child = entry['child'] ? [entry['child']] : [];
        process([...fields, ...child]);
        result.push(entry);
    });
    process(arr);
    return result;
};

const NEVER_VALUE$4 = Symbol('never');
/**
 * Counts the occurrences of each value emitted by the target observer.
 *
 * @template T - The type of values emitted by the target observer.
 * @param target - The target observer to count the occurrences for.
 * @returns - An observer that emits {@link ICounted} objects containing the value and count.
 */
const count = () => (target) => {
    return target
        .reduce((acm, cur) => {
        if (acm.value === cur) {
            return {
                value: cur,
                count: acm.count + 1,
            };
        }
        return {
            value: cur,
            count: 0,
        };
    }, {
        value: NEVER_VALUE$4,
        count: 0,
    });
};

const NEVER_VALUE$3 = Symbol('never');
/**
 * Filters out duplicate values in an observable stream based on a specified compare value.
 *
 * @template T - The type of the values in the observable stream.
 * @template V - The type of the compare value.
 * @param getCompareValue - A function that takes a value of type T and returns its compare value of type V.
 * @param target - The observable stream to filter.
 * @returns - An observable stream with distinct values.
 */
const distinct = (getCompareValue = (value) => value) => (target) => {
    let prevValue = NEVER_VALUE$3;
    return target
        .map((value) => ({
        value,
        compare: getCompareValue(value)
    }))
        .filter(({ compare: value }) => value !== prevValue)
        .tap(({ compare }) => prevValue = compare)
        .map(({ value }) => value);
};

/**
 * Groups the elements of an array in chunks of a specified size.
 *
 * @template T - The type of the elements in the array, or 'any' if not specified
 * @param by - The size of each group
 * @returns - A function that takes an array and returns a new array with the elements grouped by the specified size
 */
const group = (by) => (target) => {
    return target
        .reduce((acm, cur) => {
        if (acm.length === by) {
            return [cur];
        }
        else {
            return [...acm, cur];
        }
    }, [])
        .filter((acm) => acm.length === by);
};

/**
 * Creates an observer that emits a value after a specified delay.
 *
 * @param delay - The delay in milliseconds.
 * @returns - The observer that emits values after the specified delay.
 */
const fromInterval = (delay) => {
    let timeout;
    let iterationIdx = 0;
    const observer = new Observer(() => {
        if (timeout !== undefined) {
            clearTimeout(timeout);
        }
    });
    const process = () => {
        observer.emit(iterationIdx);
        iterationIdx++;
        timeout = setTimeout(process, delay);
    };
    observer[LISTEN_CONNECT](() => {
        process();
    });
    return observer;
};

/**
 * Creates an observable that emits the result of a given promise callback function.
 *
 * @param callbackfn - The callback function that returns a promise.
 * @param [fallbackfn] - The fallback function to handle errors if the promise rejects.
 * @returns - The observable observer.
 *
 * @template Data - The type of data emitted by the observer.
 */
const fromPromise = (callbackfn, fallbackfn) => {
    let isCanceled = false;
    const observer = new Observer(() => {
        isCanceled = true;
    });
    const process = async () => {
        try {
            const result = await callbackfn();
            if (!isCanceled) {
                observer.emit(result);
            }
        }
        catch (e) {
            if (fallbackfn) {
                fallbackfn(e);
            }
            else {
                throw e;
            }
        }
    };
    observer[LISTEN_CONNECT](() => {
        process();
    });
    return observer;
};

/**
 * Creates a delayed observer that emits a void value after a specified delay.
 *
 * @param delay - The delay in milliseconds before emitting the value.
 * @returns - The delayed observer instance.
 */
const fromDelay = (delay) => {
    let timeout;
    const observer = new Observer(() => {
        if (timeout !== undefined) {
            clearTimeout(timeout);
        }
    });
    const process = () => {
        timeout = undefined;
        observer.emit();
    };
    observer[LISTEN_CONNECT](() => {
        setTimeout(process, delay);
    });
    return observer;
};

/**
 * Creates an observer that emits a flattened and filtered array of data.
 *
 * @template Data - The type of data being observed.
 * @param data - The data to observe.
 * @returns - The observer that emits the flattened and filtered array of data.
 */
const fromArray = (data) => {
    const observer = new Observer(() => undefined);
    const process = () => {
        if (Array.isArray(data)) {
            data.flat(Number.POSITIVE_INFINITY).forEach((item) => {
                observer.emit(item);
            });
        }
        else {
            observer.emit(data);
        }
    };
    observer[LISTEN_CONNECT](() => {
        process();
    });
    return observer;
};

/**
 * A utility function for creating observer objects that can be chained
 * with multiple operators.
 *
 * @template Data The type of data emitted by the observer.
 *
 * @param factory A function that returns an observer.
 *
 * @returns The created observer object.
 */
const createObserver = (factory) => ({
    tap: (callbackfn) => factory().tap(callbackfn),
    debounce: (delay) => factory().debounce(delay),
    delay: (delay) => factory().delay(delay),
    repeat: (interval) => factory().repeat(interval),
    filter: (callbackfn) => factory().filter(callbackfn),
    map: (callbackfn) => factory().map(callbackfn),
    flatMap: (callbackfn) => factory().flatMap(callbackfn),
    operator: (callbackfn) => factory().operator(callbackfn),
    reduce: (callbackfn, begin) => factory().reduce(callbackfn, begin),
    mapAsync: (callbackfn, fallbackfn) => factory().mapAsync(callbackfn, fallbackfn),
    merge: (observer) => factory().merge(observer),
    split: () => factory().split(),
    connect: (callbackfn) => factory().connect(callbackfn),
    once: (callbackfn) => factory().once(callbackfn),
    unsubscribe: () => factory().unsubscribe(),
    share: () => factory().share(),
    toPromise: () => factory().toPromise(),
    toIteratorContext: () => factory().toIteratorContext(),
});

var _a;
/**
 * The Source class provides utility functions for creating and manipulating Observers.
 */
class Source {
}
_a = Source;
/**
 * Merges multiple observers into a single observer.
 *
 * @template A - The type of observer A.
 * @template B - The type of observer B.
 * @template C - The type of observer C.
 * @template D - The type of observer D.
 * @template E - The type of observer E.
 * @template F - The type of observer F.
 * @template G - The type of observer G.
 * @template H - The type of observer H.
 * @template I - The type of observer I.
 * @template J - The type of observer J.
 *
 * @param observers - An array of observers to merge.
 *
 * @returns - The merged observer.
 */
Source.merge = (observers) => {
    let root = new Subject().toObserver();
    observers.forEach((observer) => {
        if (observer) {
            root = root.merge(observer);
        }
    });
    return root;
};
/**
 * Creates a join observer that combines the values emitted by multiple Observers into a single Observable.
 *
 * @template A - The type of the value emitted by the first Observer.
 * @template B - The type of the value emitted by the second Observer.
 * @template C - The type of the value emitted by the third Observer.
 * @template D - The type of the value emitted by the fourth Observer.
 * @template E - The type of the value emitted by the fifth Observer.
 * @template F - The type of the value emitted by the sixth Observer.
 * @template G - The type of the value emitted by the seventh Observer.
 * @template H - The type of the value emitted by the eighth Observer.
 * @template I - The type of the value emitted by the ninth Observer.
 * @template J - The type of the value emitted by the tenth Observer.
 *
 * @param observers - An array of Observers to join.
 * @param options - Optional parameters for the join operation, including a buffer and a race flag.
 * @param options.buffer - An array to store the latest emitted values from each Observer. Defaults to an empty array.
 * @param options.race - A boolean flag indicating whether to emit the combined values immediately or wait for all Observers to emit a value. Defaults to false.
 *
 * @returns An Observer that emits an array of values, each value being the latest emitted value from the corresponding Observer.
 */
Source.join = (observers, { race = false, buffer = [], } = {}) => {
    let disposeRef = () => undefined;
    const observer = new Observer(() => disposeRef());
    observers = observers.filter((value) => !!value);
    buffer = [...new Array(observers.length)].map((_, idx) => buffer[idx]);
    const subscriptions = [];
    const next = () => {
        if (buffer.every((value) => value !== undefined)) {
            observer.emit([...buffer]);
            !race && buffer.fill(undefined);
        }
    };
    observer[LISTEN_CONNECT](() => {
        observers.forEach((observer, idx) => {
            if (observer) {
                const unsubscribe = observer.connect((value) => {
                    buffer[idx] = value;
                    next();
                });
                subscriptions.push(() => unsubscribe());
            }
        });
        disposeRef = compose(...subscriptions);
    });
    return observer;
};
/**
 * @typedef Unicast
 * @template Data - The type of data the observer handles.
 *
 * @property factory - A factory function to create the observer.
 * @property isUnicasted - Indicates whether the observer is unicast.
 *
 * @returns - A unicast observer instance.
 */
Source.unicast = (factory) => ({
    ...createObserver(factory),
    isUnicasted: true,
});
/**
 * Creates a multicast observer.
 *
 * @template Data - The type of data being observed.
 * @param factory - A factory function that creates the observer.
 * @returns - The multicast observer.
 */
Source.multicast = (factory) => {
    let observer;
    return {
        ...createObserver(() => {
            if (!observer) {
                observer = factory();
                observer[LISTEN_DISCONNECT](() => {
                    observer = undefined;
                });
            }
            return observer;
        }),
        getRef: () => observer,
        isMulticasted: true,
    };
};
/**
 * Creates a hot observable that emits data as it is received from the given emitter.
 *
 * @template Data The type of data emitted by the observable.
 * @param emitter The function that receives a callback to emit data. It should return a cleanup function or `undefined`.
 * @returns The observer that allows subscribing to and unsubscribing from the emitted data.
 */
Source.createHot = (emitter) => {
    let unsubscribeRef;
    const observer = new Observer(() => unsubscribeRef());
    unsubscribeRef = emitter(observer.emit) || (() => undefined);
    return observer;
};
/**
 * Creates a cold observable.
 *
 * @param emitter - The emitter function which is called when a subscriber is added.
 *                            It should return a function that is called when the subscription is unsubscribed,
 *                            or return `undefined` if no cleanup is needed.
 * @returns - The created observer.
 */
Source.createCold = (emitter) => {
    let unsubscribeRef = () => undefined;
    const observer = new Observer(() => unsubscribeRef());
    observer[LISTEN_CONNECT](() => {
        unsubscribeRef = emitter(observer.emit) || (() => undefined);
    });
    return observer;
};
/**
 * Creates a new instance of the Cold object.
 */
Source.create = _a.createCold;
/**
 * Creates a pipe that connects an observer to a subject and emits output values based on a given emitter function.
 *
 * @param target - The observer that will receive output values.
 * @param emitter - A function that takes a subject and a next function and returns an unsubscribe function.
 * @returns The observer that is connected to the subject and emits output values.
 * @template Data - The type of data that will be observed.
 * @template Output - The type of output that will be emitted.
 */
Source.pipe = (target, emitter) => {
    let unsubscribeRef = () => undefined;
    const observer = new Observer(() => unsubscribeRef());
    observer[LISTEN_CONNECT](() => {
        const subject = new Subject();
        const unsubscribeTarget = target.connect(subject.next);
        const unsubscribeEmitter = emitter(subject, observer.emit) || (() => undefined);
        unsubscribeRef = compose(() => unsubscribeEmitter(), () => unsubscribeTarget());
    });
    return observer;
};
Source.fromInterval = fromInterval;
Source.fromPromise = fromPromise;
Source.fromDelay = fromDelay;
Source.fromArray = fromArray;
/**
 * Creates a new observer that emits a value from the given data or function.
 *
 * @param data - The data or function to emit from the observer.
 * @returns - The created observer.
 */
Source.fromValue = (data) => {
    const observer = new Observer(() => undefined);
    observer[LISTEN_CONNECT](() => {
        if (typeof data === 'function') {
            observer.emit(data());
        }
        else {
            observer.emit(data);
        }
    });
    return observer;
};
/**
 * Creates an observer from the given subject and returns it.
 *
 * @template Data - The type of data emitted by the observer.
 * @param subject - The subject to create the observer from.
 * @returns - The observer created from the subject.
 */
Source.fromSubject = (subject) => {
    let unsubscribeRef;
    const observer = new Observer(() => unsubscribeRef());
    unsubscribeRef = subject.subscribe(observer.emit);
    return observer;
};
/**
 * Creates an observer from a BehaviorSubject.
 *
 * @template Data The type of data emitted by the BehaviorSubject.
 * @param subject - The BehaviorSubject to create the observer from.
 * @returns The observer created from the BehaviorSubject.
 */
Source.fromBehaviorSubject = (subject) => {
    let unsubscribeRef;
    const observer = new Observer(() => unsubscribeRef());
    observer[LISTEN_CONNECT](() => {
        subject.data && observer.emit(subject.data);
    });
    unsubscribeRef = subject.subscribe(observer.emit);
    return observer;
};
/*
Source.join([
    Source.create<string>((next) => next("1")),
    Source.create<number>((next) => next(2)),
    Source.create<boolean>((next) => next(false)),
]).split().connect((value) => console.log(value));
*/
/*
const { Source } = require('.')
const multicast = Source.multicast(() => Source.create(() => {
    console.log('ctor');
    return () => console.log('dtor');
}));
const c1 = multicast.connect((v) => console.log(v))
const c2 = multicast.connect((v) => console.log(v))
c1()
c2()
const c3 = multicast.connect((v) => console.log(v))
*/

const NEVER_VALUE$2 = Symbol('never');
/**
 * Takes a count number as an input and returns a higher-order function
 * that accepts a target TObserver. The higher-order function returns a TObserver
 * that filters the values emitted by the target TObserver based on the count.
 * Only values with an index less than or equal to the count will be emitted.
 *
 * @typeparam T - The type of values emitted by the target TObserver.
 * @param count - The maximum index of values to be emitted.
 * @returns A higher-order function that accepts a target TObserver and returns
 * a TObserver that filters values based on the count.
 */
const take = (count) => (target) => {
    return target
        .reduce((acm, cur) => ({
        value: cur,
        idx: acm.idx + 1,
    }), {
        value: NEVER_VALUE$2,
        idx: 0
    })
        .filter(({ idx }) => idx <= count)
        .map(({ value }) => value);
};

const NEVER_VALUE$1 = Symbol('never');
/**
 * Adds liveness functionality to a given observer stream.
 *
 * @template T - The type of values in the observer stream.
 * @param fallbackfn - The fallback function to be called when liveness times out.
 * @param waitFor - The time period in milliseconds to wait for liveness before calling the fallback function. Default is 5000 milliseconds.
 * @param target - The observer stream to add liveness functionality to.
 * @returns - The modified observer stream with liveness functionality.
 */
const liveness = (fallbackfn, waitFor = 5000) => (target) => {
    let isOk = true;
    const stream = Source.merge([
        Source.fromInterval(waitFor)
            .filter(() => {
            const isOkPrev = isOk;
            isOk = false;
            return !isOkPrev;
        })
            .operator(take(1))
            .tap(fallbackfn)
            .map(() => NEVER_VALUE$1),
        target.tap(() => {
            isOk = true;
        })
    ]);
    return stream.filter((value) => value !== NEVER_VALUE$1);
};

/**
 * Applies a pairwise transformation to an observer.
 *
 * @template T - The type of value observed by the original observer.
 * @param [by=2] - The number of values in each pair.
 * @returns - A function that takes an observer and returns a new observer that emits pairs of values.
 */
const pair = (by = 2) => (target) => {
    return target
        .reduce((acm, cur) => [...acm, cur].slice(-by), [])
        .filter((acm) => acm.length === by);
};

const NEVER_VALUE = Symbol('never');
/**
 * Skips the first n elements emitted by an observer.
 *
 * @typeparam T - The type of elements emitted by the observer.
 * @param the - The number of elements to skip.
 * @returns A new observer that skips the first n elements of the original observer.
 */
const skip = (the) => (target) => {
    return target
        .reduce((acm, cur) => ({
        value: cur,
        idx: acm.idx + 1,
    }), {
        value: NEVER_VALUE,
        idx: 0
    })
        .filter(({ idx }) => idx > the)
        .map(({ value }) => value);
};

/**
 * Applies stride tricks to a given target observer.
 * @template T The type of elements in the target observer.
 * @param strideSize The size of each stride.
 * @param [step=Math.floor(strideSize / 2)] The step size between each stride.
 * @returns The transformed observer that emits strided data.
 * @throws If the strideSize or step is too big, or if the data is unshaped.
 */
const strideTricks = (strideSize, step = Math.floor(strideSize / 2)) => (target) => {
    let windowSize = -1;
    let totalSteps = -1;
    let resultSize = -1;
    let needExtraStep = false;
    return target
        .tap((buffer) => {
        if (windowSize === -1) {
            windowSize = buffer.length;
            totalSteps = Math.ceil((windowSize - strideSize) / step);
            needExtraStep = totalSteps * strideSize !== windowSize;
            resultSize = totalSteps + (needExtraStep ? 1 : 0);
            if (strideSize > windowSize || step > strideSize)
                throw new Error('rn-declarative strideTricks too big stride');
        }
    })
        .flatMap((buffer) => {
        if (buffer.length !== windowSize) {
            throw new Error('rn-declarative strideTricks unshaped data');
        }
        const strides = [];
        for (let i = 0; i !== totalSteps; i++) {
            const startPos = i * step;
            strides.push(buffer.slice(startPos, startPos + strideSize));
        }
        if (needExtraStep) {
            const lastStep = buffer.slice(windowSize - strideSize, windowSize);
            lastStep["lastStep"] = true;
            strides.push(lastStep);
        }
        return strides.flat(1);
    })
        .reduce((acm, cur) => {
        if (acm.length === strideSize) {
            return [cur];
        }
        else {
            return [...acm, cur];
        }
    }, [])
        .filter((acm) => acm.length === strideSize)
        .reduce((acm, cur) => {
        if (acm.length === resultSize) {
            return [cur];
        }
        else {
            return [...acm, cur];
        }
    }, [])
        .filter((acm) => acm.length === resultSize);
};

/**
 * Represents a collection of static operator functions.
 */
class Operator {
}
Operator.take = take;
Operator.skip = skip;
Operator.pair = pair;
Operator.group = group;
Operator.strideTricks = strideTricks;
Operator.distinct = distinct;
Operator.liveness = liveness;
Operator.count = count;
/*
const { Source, Operator } = require('.')

Source.fromInterval(1_000).operator(Operator.take(5)).connect(console.log)

Source.fromInterval(1_000).operator(Operator.pair()).connect(console.log)

Source.fromInterval(1_000).operator(Operator.group(3)).connect(console.log)

Source.fromInterval(1_000).operator(Operator.skip(3)).connect(console.log)

Source.fromValue([1,2,3,4,5,6,7,8,9,10,11,12,13]).operator(Operator.strideTricks(3, 3)).connect(console.log)

*/

/**
 * A function that picks a subset of documents from an array of documents, given a limit and offset.
 *
 * @template T - The type of the documents in the array.
 * @param limit - The maximum number of documents to pick.
 * @param offset - The number of documents to skip before picking.
 * @returns - A function that takes an array of documents and returns an object with `rows` and `done` properties.
 *                       The `rows` property contains the picked documents, and `done` property indicates if the picking is finished.
 */
const pickDocuments = (limit, offset) => {
    const result = [];
    return (rows = []) => {
        for (const row of rows) {
            if (offset > 0) {
                offset -= 1;
                continue;
            }
            if (limit > 0) {
                result.push(row);
                limit -= 1;
                continue;
            }
            break;
        }
        return {
            rows: result,
            done: limit <= 0,
        };
    };
};

/**
 * Resolves the documents from an async generator and paginates them.
 *
 * @param iterator - The async generator to resolve documents from.
 * @returns - A promise that resolves to the flattened array of documents.
 */
const paginateDocuments = async (iterator, limit, offset) => {
    const iter = pickDocuments(limit, offset);
    for await (const chunk of iterator) {
        const rows = [chunk].flatMap(v => v);
        if (iter(rows).done) {
            break;
        }
    }
    return iter().rows;
};

/**
 * Resolves the documents from an async generator and distincts them.
 *
 * @param iterator - The async generator to resolve documents from.
 * @returns - A promise that resolves to the flattened array of documents.
 */
async function* distinctDocuments(iterator, getId = (data) => data["id"]) {
    const duplicateSet = new Set();
    for await (const chunk of iterator) {
        const rows = [chunk].flatMap(v => v);
        for (const row of rows) {
            const id = getId(row);
            if (!duplicateSet.has(id)) {
                duplicateSet.add(id);
                yield row;
            }
        }
    }
}

/**
 * Resolves the documents from an async generator and returns them as a flattened array.
 *
 * @param iterator - The async generator to resolve documents from.
 * @returns - A promise that resolves to the flattened array of documents.
 */
const resolveDocuments = async (iterator) => {
    const result = [];
    for await (const item of iterator) {
        result.push(item);
    }
    return result.flatMap(v => v);
};

/**
 * Filters the documents from an async generator and yield them by the condition
 *
 * @param iterator - The async generator to resolve documents from.
 * @returns - A promise that resolves to the flattened array of documents.
 */
async function* filterDocuments(iterator, predicate) {
    for await (const chunk of iterator) {
        const rows = [chunk].flatMap(v => v);
        for (const row of rows) {
            if (await predicate(row)) {
                yield row;
            }
        }
    }
}

/**
 * Maps the documents from an async generator and yield them
 *
 * @param iterator - The async generator to resolve documents from.
 * @returns - A promise that resolves to the flattened array of documents.
 */
async function* mapDocuments(iterator, callbackfn) {
    for await (const chunk of iterator) {
        const rows = [chunk].flatMap(v => v);
        for (const row of rows) {
            yield await callbackfn(row);
        }
    }
}

const TOTAL_DOCUMENTS = 10000;
const REQUEST_LIMIT = 5000;
const REQUEST_DELAY = 100;
/**
 * Asynchronous generator function that iterates over documents.
 *
 * @template Data - The type of the row data in the documents.
 *
 * @param config - The configuration object.
 * @param [config.totalDocuments=TOTAL_DOCUMENTS] - The total number of documents to iterate over.
 * @param [config.limit=REQUEST_LIMIT] - The number of documents to fetch in each request.
 * @param [config.delay=REQUEST_DELAY] - The delay between each request.
 * @param [config.createRequest=() => []] - The function used to create the request.
 *
 * @returns An asynchronous generator that yields an array of documents.
 *
 * @throws If the response length is greater than the specified limit.
 */
const iterateDocuments = async function* ({ totalDocuments = TOTAL_DOCUMENTS, limit = REQUEST_LIMIT, delay = REQUEST_DELAY, getId = (data) => data["id"], createRequest = () => [], }) {
    /**
     * Creates a request and returns the result asynchronously.
     *
     * @name createRequest
     * @function
     * @param args - The arguments to pass to the createRequest function.
     * @returns A promise that resolves to the result of the request.
     */
    const request = async (...args) => {
        const [result] = await Promise.all([
            createRequest(...args),
            sleep(delay),
        ]);
        return result;
    };
    let counter = 0;
    let lastId = null;
    /**
     * Represents the last query made to the server.
     *
     * @typedef LastQuery
     * @property lastId - The ID of the last object fetched. Defaults to null.
     * @property offset - The offset used in pagination. Defaults to 0.
     * @property page - The current page number. Defaults to 0.
     * @property limit - The maximum number of objects to fetch per request.
     */
    let lastQuery = request({
        lastId: null,
        offset: 0,
        page: 0,
        limit,
    });
    while (counter < totalDocuments) {
        const response = await lastQuery;
        if (response.length < limit) {
            yield response;
            break;
        }
        if (response.length > limit) {
            throw new Error('functool-kit iterateDocuments response.length > limit');
        }
        lastId = getId(response[response.length - 1]) || null;
        counter += limit;
        /**
         * Represents the last query made by the user.
         * @class
         */
        lastQuery = request({
            lastId,
            offset: counter,
            page: Math.ceil(counter / limit),
            limit,
        });
        yield response;
    }
};

async function* iteratePromise(fn) {
    for (const row of await fn()) {
        yield row;
    }
}

const iterateUnion = (iterators, getId = (data) => data["id"]) => async function* (limit, offset) {
    const duplicateSet = new Set();
    for (const iterator of iterators) {
        for await (const chunk of iterator) {
            const rows = [chunk].flatMap(v => v);
            for (const row of rows) {
                const id = getId(row);
                if (duplicateSet.has(id)) {
                    continue;
                }
                if (offset > 0) {
                    offset -= 1;
                    continue;
                }
                if (limit > 0) {
                    duplicateSet.add(id);
                    yield row;
                    limit -= 1;
                    continue;
                }
                return;
            }
        }
    }
};

async function* iterateList(rows, map = (row) => Promise.resolve(row)) {
    for (const row of rows) {
        yield await map(row);
    }
}

const has = (arr, ...value) => {
    if (Array.isArray(arr)) {
        return value.some((v) => arr.includes(v));
    }
    if (arr instanceof Set) {
        return value.some((v) => arr.has(v));
    }
    if (arr instanceof Map) {
        return value.some((v) => arr.has(v));
    }
    if (arr) {
        return value.some((v) => arr === v);
    }
    return false;
};

/**
 * Performs a logical AND operation on multiple values.
 *
 * @template T - The type of the values
 * @param args - The values to perform the logical AND operation on
 * @returns - The result of the logical AND operation
 */
const and = (...args) => {
    if (args.some((arg) => arg instanceof Promise)) {
        return new Promise(async (res, rej) => {
            try {
                const items = await Promise.all(args);
                const result = items.reduce((acm, cur) => Boolean(acm && cur), true);
                res(result);
            }
            catch (error) {
                rej(error);
            }
        });
    }
    return args.reduce((acm, cur) => Boolean(acm && cur), true);
};

/**
 * Returns a value of type T representing the logical OR operation on the given arguments.
 *
 * @param args - The arguments to be evaluated for the logical OR operation.
 * @returns A value of type T representing the result of the logical OR operation.
 * @throws If any of the arguments is a rejected promise.
 * @typeparam T - The type of the arguments and the return value.
 */
const or = (...args) => {
    if (args.some((arg) => arg instanceof Promise)) {
        return new Promise(async (res, rej) => {
            try {
                const items = await Promise.all(args);
                const result = items.reduce((acm, cur) => Boolean(acm || cur), false);
                res(result);
            }
            catch (error) {
                rej(error);
            }
        });
    }
    return args.reduce((acm, cur) => Boolean(acm || cur), false);
};

/**
 * Applies the logical negation operator to the given argument.
 * If the argument is a Promise, it returns a new Promise that resolves to the negation of the resolved value of the argument Promise.
 * If the argument is not a Promise, it returns the negation of the argument.
 *
 * @template T - The type of the argument and the return value.
 * @param arg - The argument to apply the logical negation operator.
 * @returns - The result of apply the logical negation operator to the argument.
 */
const not = (arg) => {
    if (arg instanceof Promise) {
        return new Promise(async (res, rej) => {
            try {
                const result = await arg;
                res(!result);
            }
            catch (error) {
                rej(error);
            }
        });
    }
    return !arg;
};

/**
 * Executes conditional branching based on the provided conditions and returns the appropriate value.
 *
 * @template A - Type of the condition value.
 * @template T - Type of the run value.
 * @template E - Type of the not value.
 *
 * @param params - The parameters object.
 * @param params.condition - The condition value or a function that returns the condition value.
 * @param params.run - The run value or a function that returns the run value.
 * @param [params.not=false] - The not value or a function that returns the not value. Defaults to false.
 *
 * @returns - The result of executing the condition and returning the appropriate value.
 */
const match = ({ condition, run, not = false }) => {
    const check = typeof condition === 'function' ? condition() : condition;
    const result = typeof run === 'function' ? run() : run;
    const fallback = typeof not === 'function' ? not : not;
    if (result instanceof Promise || check instanceof Promise || fallback instanceof Promise) {
        return new Promise(async (res, rej) => {
            try {
                if (await check) {
                    res(await result);
                    return;
                }
                res(await fallback);
            }
            catch (error) {
                rej(error);
            }
        });
    }
    if (check) {
        return result;
    }
    return fallback;
};

/**
 * Joins multiple arrays into a single array, removing duplicates and filtering out null values.
 *
 * @param arr - The arrays to join.
 * @returns - The joined array.
 *
 * @template T - The type of values in the array.
 */
const join = (...arr) => {
    arr = arr.flat(1);
    return [...new Set(arr.flatMap((item) => {
            if (Array.isArray(item)) {
                return join(item);
            }
            return item;
        }).filter(Boolean))];
};

/**
 * Joins an array of strings or arrays of strings into a single string, with each element separated by a space.
 * It also handles cases where the input is null or nested arrays.
 */
const str = (...arr) => {
    return join(...arr).join(" ");
};
/**
 * Joins an array of strings or arrays of strings into a single string, with each element separated by a newline.
 */
str.newline = (...arr) => {
    return join(...arr).join("\n");
};
/**
 * Joins an array of strings or arrays of strings into a single string, with each element separated by a space.
 */
str.space = (...arr) => {
    return join(...arr).join(" ");
};
/**
 * Joins an array of strings or arrays of strings into a single string, with each element separated by a comma.
 */
str.comma = (...arr) => {
    return join(...arr).join(", ");
};
/**
 * Joins an array of strings or arrays of strings into a single string, with each element separated by a period.
 */
str.dot = (...arr) => {
    return join(...arr).join(". ");
};
/**
 * Joins an array of strings or arrays of strings into a single string, with each element separated by a semicolon.
 */
str.semicolon = (...arr) => {
    return join(...arr).join(";");
};

/**
 * Returns the last element of an array or null if the array is empty or not an array.
 *
 * @param arr - The array from which to retrieve the last element.
 * @returns - The last element of the array or null.
 */
const last = (arr) => {
    if (Array.isArray(arr)) {
        const [last] = arr.slice(-1);
        return last || null;
    }
    return null;
};

/**
 * Filters an array and removes null values, casting the resulting array to a specific generic type if specified.
 *
 * @param arr - The input array containing values of type T or null.
 * @returns - The filtered array with null values removed.
 *
 * @template T - The generic type of the array elements.
 */
const truely = (arr) => {
    return (arr?.filter(v => v) || []);
};

const SPLIT_CHARS = ["_", "-", " "];
const split = (...arr) => {
    return join(...arr)
        .flatMap((c) => c)
        .flatMap((c) => {
        const separator = SPLIT_CHARS.find((s) => c.includes(s));
        if (separator) {
            return c.split(separator).map((c) => c.toLowerCase());
        }
        return c.toLowerCase();
    });
};

/**
 * Extracts the enumerable properties from an Error object and returns them in a new object.
 *
 * @param error - The Error object.
 * @returns - An object containing the enumerable properties of the Error object.
 */
const errorData = (error) => {
    if (!error) {
        return {};
    }
    const propertyNames = Object.getOwnPropertyNames(error);
    const result = {};
    propertyNames.forEach((property) => {
        const descriptor = Object.getOwnPropertyDescriptor(error, property) || {};
        if ('value' in descriptor) {
            result[property] = descriptor.value;
        }
    });
    return result;
};

const FETCH_CONFIG = {
    useSymbolException: false,
};
/**
 * Represents an error that occurs during a fetch request.
 *
 * @class
 * @extends Error
 */
class FetchError extends Error {
    constructor(originalError, request, response, statusCode, info, init) {
        super(originalError.message || 'FetchError');
        this.originalError = originalError;
        this.request = request;
        this.response = response;
        this.statusCode = statusCode;
        this.info = info;
        this.init = init;
    }
}
const PAYLOAD_METHODS = ['POST', 'PUT', 'PATCH'];
/**
 * Makes an asynchronous HTTP request using the Fetch API.
 *
 * @param input - The resource URL or an instance of the URL class.
 * @param [init] - The request options.
 * @returns - The response data as a Promise.
 * @throws - If an error occurs during the request.
 */
const fetchApi = async (input, init) => {
    const request = input instanceof URL ? input.toString() : input;
    let response = undefined;
    try {
        response = await fetch(request, {
            ...init,
            headers: {
                ...(PAYLOAD_METHODS.includes(init?.method?.toUpperCase()) && {
                    "Content-Type": "application/json",
                }),
                ...init?.headers,
            },
        });
        if (!response.ok) {
            const responseText = await trycatch(async () => await response.text(), { defaultValue: null })();
            const requestInfo = trycatch(JSON.stringify, { defaultValue: null })(input);
            const requestInit = trycatch(JSON.stringify, { defaultValue: null })(init);
            throw new Error(`fetchApi response not ok. Info: ${requestInfo}, Init: ${requestInit}, Status: ${response.status}, Message: ${responseText}`);
        }
        return await response.json();
    }
    catch (error) {
        if (FETCH_CONFIG.useSymbolException) {
            return CATCH_SYMBOL;
        }
        throw new FetchError(error, request, response, response?.status || 0, input || null, init || null);
    }
};
fetchApi.config = (config) => {
    Object.assign(FETCH_CONFIG, config);
};

class ToolRegistry {
    constructor(registryName, tools = {}) {
        this.registryName = registryName;
        this.tools = tools;
        this.register = (name, tool) => {
            if (name in this.tools) {
                throw new Error(`functools-kit Tool is already registered name=${String(name)} registryName=${this.registryName} (register)`);
            }
            return new ToolRegistry(this.registryName, {
                ...this.tools,
                [name]: tool,
            });
        };
        this.override = (name, tool) => {
            if (name in this.tools) {
                return new ToolRegistry(this.registryName, {
                    ...this.tools,
                    [name]: isObject(tool)
                        ? Object.assign({}, this.tools[name], tool)
                        : tool,
                });
            }
            return new ToolRegistry(this.registryName, {
                ...this.tools,
                [name]: tool,
            });
        };
        this.get = (name) => {
            if (name in this.tools) {
                return this.tools[name];
            }
            throw new Error(`functools-kit Tool not registered name=${String(name)} registryName=${this.registryName} (get)`);
        };
        this.init = () => {
            for (const tool of Object.values(this.tools)) {
                // @ts-ignore
                tool.init && tool.init();
            }
        };
    }
}

class SortedArray {
    constructor() {
        this.items = [];
    }
    push(item, score) {
        const newEntry = { item, score };
        let left = 0;
        let right = this.items.length - 1;
        while (left <= right) {
            const mid = Math.floor((left + right) / 2);
            if (this.items[mid].score === score) {
                left = mid;
                break;
            }
            else if (this.items[mid].score > score) {
                left = mid + 1;
            }
            else {
                right = mid - 1;
            }
        }
        this.items.splice(left, 0, newEntry);
    }
    pop(item) {
        const index = this.items.findIndex((entry) => entry.item === item);
        if (index !== -1) {
            this.items.splice(index, 1);
            return true;
        }
        return false;
    }
    getItems() {
        return this.items.map((entry) => entry.item);
    }
    getEntries() {
        return this.items;
    }
    take(n, minScore = Number.POSITIVE_INFINITY) {
        const result = [];
        let count = 0;
        for (const { item, score } of this.items) {
            if (score < minScore) {
                break;
            }
            if (count >= n) {
                break;
            }
            result.push(item);
            count++;
        }
        return result;
    }
    [Symbol.iterator]() {
        let index = 0;
        const items = this.items || [];
        return {
            next: () => {
                if (index < items.length) {
                    return { value: items[index++].item, done: false };
                }
                else {
                    return { done: true };
                }
            },
        };
    }
    get length() {
        return this.items.length;
    }
}

class LimitedSet extends Set {
    constructor(_maxSize = 20) {
        super();
        this._maxSize = _maxSize;
    }
    add(value) {
        if (this.size >= this._maxSize) {
            const lastElement = Array.from(this).shift();
            this.delete(lastElement);
        }
        return super.add(value);
    }
}

class LimitedMap extends Map {
    constructor(_maxSize = 20) {
        super();
        this._maxSize = _maxSize;
    }
    set(key, value) {
        if (this.size >= this._maxSize) {
            const lastKey = Array.from(this.keys()).shift();
            this.delete(lastKey);
        }
        return super.set(key, value);
    }
}

exports.BehaviorSubject = BehaviorSubject;
exports.CANCELED_PROMISE_SYMBOL = CANCELED_PROMISE_SYMBOL;
exports.CATCH_SYMBOL = CATCH_SYMBOL;
exports.EventEmitter = EventEmitter;
exports.FetchError = FetchError;
exports.LimitedMap = LimitedMap;
exports.LimitedSet = LimitedSet;
exports.Observer = Observer;
exports.Operator = Operator;
exports.PubsubArrayAdapter = PubsubArrayAdapter;
exports.PubsubMapAdapter = PubsubMapAdapter;
exports.RateError = RateError;
exports.SortedArray = SortedArray;
exports.Source = Source;
exports.Subject = Subject;
exports.TIMEOUT_SYMBOL = TIMEOUT_SYMBOL;
exports.Task = Task;
exports.ToolRegistry = ToolRegistry;
exports.afterinit = afterinit;
exports.and = and;
exports.cached = cached;
exports.cancelable = cancelable;
exports.compareArray = compareArray;
exports.compareFulltext = compareFulltext;
exports.compose = compose;
exports.createAwaiter = createAwaiter;
exports.debounce = debounce;
exports.deepFlat = deepFlat;
exports.distinctDocuments = distinctDocuments;
exports.errorData = errorData;
exports.execpool = execpool;
exports.fetchApi = fetchApi;
exports.filterDocuments = filterDocuments;
exports.first = first;
exports.formatText = formatText;
exports.getErrorMessage = getErrorMessage;
exports.has = has;
exports.isObject = isObject;
exports.iterateDocuments = iterateDocuments;
exports.iterateList = iterateList;
exports.iteratePromise = iteratePromise;
exports.iterateUnion = iterateUnion;
exports.join = join;
exports.last = last;
exports.lock = lock;
exports.makeExtendable = makeExtendable;
exports.mapDocuments = mapDocuments;
exports.match = match;
exports.memoize = memoize;
exports.not = not;
exports.obsolete = obsolete;
exports.or = or;
exports.paginateDocuments = paginateDocuments;
exports.pickDocuments = pickDocuments;
exports.pubsub = pubsub;
exports.queued = queued;
exports.randomString = randomString;
exports.rate = rate;
exports.resolveDocuments = resolveDocuments;
exports.retry = retry;
exports.schedule = schedule;
exports.singlerun = singlerun;
exports.singleshot = singleshot;
exports.singletick = singletick;
exports.sleep = sleep;
exports.split = split;
exports.str = str;
exports.throttle = throttle;
exports.timeout = timeout;
exports.truely = truely;
exports.trycatch = trycatch;
exports.ttl = ttl;
exports.waitForNext = waitForNext;
